"""
gds_fdtd simulation toolbox.

Layout processing module.
@author: Mustafa Hammood, 2025
"""

from .core import layout, port, structure, region
import logging
import klayout.db as pya


def dilate(vertices, extension=1.3):
    """grow or shrink a rectangle defined as [[x1,y1],[x2,y2]]

    Args:
        vertices (list): list defining rectangle: [[x1,y1],[x2,y2]]
        extension (int, optional): Growth amount. Defaults to 1.

    Returns:
        list: dilated rectangle.
    """
    import numpy as np

    x_min = np.min([i[0] for i in vertices]) - extension
    x_max = np.max([i[0] for i in vertices]) + extension
    y_min = np.min([i[1] for i in vertices]) - extension
    y_max = np.max([i[1] for i in vertices]) + extension

    return [[x_min, y_min], [x_max, y_min], [x_max, y_max], [x_min, y_max]]


def dilate_1d(vertices, extension=1, dim="y"):
    if dim == "x":
        if vertices[0][0] < vertices[1][0]:
            sign = 1
        else:
            sign = -1
        return [
            [vertices[0][0] - abs(extension) * sign, vertices[0][1]],
            [vertices[1][0] + abs(extension) * sign, vertices[1][1]],
        ]
    elif dim == "y":
        if vertices[0][1] < vertices[1][1]:
            sign = 1
        else:
            sign = -1
        return [
            [vertices[0][0], vertices[0][1] - abs(extension)] * sign,
            [vertices[1][0], vertices[1][1] + abs(extension) * sign],
        ]
    elif dim == "xy":
        if vertices[0][1] < vertices[1][1]:
            sign = 1
        else:
            sign = -1
        return [
            [vertices[0][0], vertices[0][1] - abs(extension)] * sign,
            [
                vertices[1][0] + abs(extension) * sign,
                vertices[1][1] + abs(extension) * sign,
            ],
        ]

    else:
        raise ValueError("Dimension must be 'x' or 'y' or 'xy'")


def apply_prefab(fname, top_cell, MODEL_NAME="ANT_NanoSOI_ANF1_d9"):
    import prefab as pf

    device = pf.read.from_gds(gds_path=fname, cell_name=top_cell)

    prediction = device.predict(model=pf.models[MODEL_NAME])
    prediction_bin = prediction.binarize()

    prediction_bin.to_gds(gds_path=fname, cell_name=top_cell, gds_layer=(1, 0))
    return


def load_device(
    fname: str,
    tech,
    top_cell: str = None,
    z_span: float = 3.0,
    z_center: float | None = None,
    prefab=None,
):
    from .simprocessor import load_component_from_tech

    ly = pya.Layout()
    ly.read(fname)

    if top_cell is None:
        if len(ly.top_cells()) > 1:
            err_msg = "More than one top cell found, ensure only 1 top cell exists. Otherwise, specify the cell using the top_cell argument."
            logging.error(err_msg)
            raise ValueError(err_msg)
        else:
            cell = ly.top_cell()
            name = cell.name
    else:
        cell = ly.cell(top_cell)
        if cell is None:
            err_msg = f"Top cell with name {top_cell} not found."
            logging.error(err_msg)
            raise ValueError(err_msg)
        name = cell.name

    c = load_component_from_tech(
        ly=layout(name, ly, cell), tech=tech, z_span=z_span, z_center=z_center
    )

    dbu = ly.dbu  # Get the database unit (dbu) from the layout

    for p in c.ports:
        polygon = p.polygon_extension(buffer=2)
        layer_index = ly.layer(pya.LayerInfo(1, 0))
        # Convert polygon vertices from um to dbu
        polygon_dbu = [pya.Point(int(pt[0] / dbu), int(pt[1] / dbu)) for pt in polygon]
        cell.shapes(layer_index).insert(pya.Polygon(polygon_dbu))

    new_layout_path = fname.replace(".gds", "_with_extensions.gds")
    ly.write(new_layout_path)

    if prefab is not None:
        apply_prefab(fname=new_layout_path, top_cell=top_cell, MODEL_NAME=prefab)


def load_cell(fname: str, top_cell: str = None) -> pya.Cell:
    """
    Load a GDS layout and return a cell object.

    Args:
        fname (str): Path to the GDS file.
        top_cell (str, optional): Name of the top cell. If None, the function will attempt to find a single top cell. Defaults to None.

    Returns:
        pya.Cell: A cell object containing the name, layout, and top cell.

    Raises:
        ValueError: If more than one top cell is found and top_cell is not specified, or if the specified top cell is not found.
    """
    ly = pya.Layout()
    ly.read(fname)

    if top_cell is None:
        if len(ly.top_cells()) > 1:
            err_msg = "More than one top cell found, ensure only 1 top cell exists. Otherwise, specify the cell using the top_cell argument."
            logging.error(err_msg)
            raise ValueError(err_msg)
        else:
            cell = ly.top_cell()
            name = cell.name
    else:
        cell = ly.cell(top_cell)
        if cell is None:
            err_msg = f"Top cell with name {top_cell} not found."
            logging.error(err_msg)
            raise ValueError(err_msg)
        name = cell.name

    return cell, ly


def load_region(
    cell: pya.Cell,
    layer: list[int, int] = [68, 0],
    z_center: float = 0.0,
    z_span: float = 5.0,
    extension: float = 1.3,
):
    """
    Get device bounds.

    Args:
        cell (pya.Cell): SiEPIC Tidy3d cell type to extract the polygons from.
        layer (list[int, int]): Layer to detect the devrec object from. Defaults to [68, 0].
        z_center (float): Z-center of the layout in microns. Defaults to 0.
        z_span (float): Z-span of the layout in microns. Defaults to 5.
        extension (float): Amount of extended region to retrieve beyond the specified region. Defaults to 1.3.

    Returns:
        region: Region object type.
    """

    dbu = cell.layout().dbu
    layer = cell.layout().layer(layer[0], layer[1])
    iter1 = cell.begin_shapes_rec(layer)
    # DevRec must be either a Box or a Polygon:
    if iter1.shape().is_box():
        box = iter1.shape().box.transformed(iter1.itrans())
        polygon = pya.Polygon(box)  # Save the component outline polygon
        DevRec_polygon = pya.Polygon(iter1.shape().box)
    if iter1.shape().is_polygon():
        polygon = iter1.shape().polygon.transformed(
            iter1.itrans()
        )  # Save the component outline polygon
        DevRec_polygon = iter1.shape().polygon
    polygons_vertices = [
        [[vertex.x * dbu, vertex.y * dbu] for vertex in p.each_point()]
        for p in [p.to_simple_polygon() for p in [DevRec_polygon]]
    ][0]

    if extension != 0:
        polygons_vertices = dilate(polygons_vertices, extension)
    return region(vertices=polygons_vertices, z_center=z_center, z_span=z_span)


def load_structure(cell, name, layer, z_base, z_span, material, sidewall_angle=90):
    """
    Extract polygons from a given cell on a given layer.

    Parameters
    ----------
    cell : klayout.db (pya) Cell type
        Cell to extract the polygons from.
    layer : klayout.db (pya) layout.layer() type
        Layer to place the pin object into.
    dbu : Float, optional
        Layout's database unit (in microns). The default is 0.001 (1 nm)

    Returns
    -------
    polygons_vertices : list [lists[x,y]]
        list of polygons from the cell.

    """

    dbu = cell.layout().dbu
    layer = cell.layout().layer(layer[0], layer[1])

    r = pya.Region()
    s = cell.begin_shapes_rec(layer)
    while not (s.at_end()):
        if s.shape().is_polygon() or s.shape().is_box() or s.shape().is_path():
            r.insert(s.shape().polygon.transformed(s.itrans()))
        s.next()

    r.merge()
    polygons = [p for p in r.each_merged()]
    polygons_vertices = [
        [[vertex.x * dbu, vertex.y * dbu] for vertex in p.each_point()]
        for p in [p.to_simple_polygon() for p in polygons]
    ]
    structures = []
    for idx, s in enumerate(polygons_vertices):
        name = f"{name}_{idx}"
        structures.append(
            structure(
                name=name,
                polygon=s,
                z_base=z_base,
                z_span=z_span,
                material=material,
                sidewall_angle=sidewall_angle,
            )
        )
    return structures


def load_structure_from_bounds(bounds, name, z_base, z_span, material, extension=2.0):
    """Load a structure from a region definition

    Args:
        bounds (core.region): Input region to use to generate structure.
        name (_type_): Name of structure.
        z_base (float): Z base of structure.
        z_span (float): Z span (thickness) of structure, can be negative for downward growth.
        material (tidy3d.Medium): Material of structure
        extension (float, optional): Growth (or shrinkage), in um, of structure defintion relative to bounds. Defaults to 2 um.

    Returns:
        core.structure: Structure generated from input region.
    """
    return structure(
        name=name,
        polygon=dilate(bounds.vertices, extension=extension),
        z_base=z_base,
        z_span=z_span,
        material=material,
    )


def load_ports(cell: pya.Cell, layer: list[int, int] = [1, 10]):
    """Load ports from cell.

    Args:
        cell (pya.Cell): Input cell object
        layer (list, optional): Ports layer identifier. Defaults to [1, 10].

    Returns:
        list: List of extracted port objects.
    """
    def get_direction(path):
        """Determine orientation of a pin path."""
        if path.points > 2:
            return ValueError("Number of points in a pin path are > 2.")
        p = path.each_point()
        p1 = p.__next__()
        p2 = p.__next__()
        if p1.x == p2.x:
            if p1.y > p2.y:  # north/south
                return 270
            else:
                return 90
        elif p1.y == p2.y:  # east/west
            if p1.x > p2.x:
                return 180
            else:
                return 0

    def get_center(path, dbu):
        """Determine center of a pin path."""
        p = path.each_point()
        p1 = p.__next__()
        p2 = p.__next__()
        direction = get_direction(path)
        if direction in [0, 180]:
            x = dbu * (p1.x + p2.x) / 2
            y = dbu * p1.y
        elif direction in [90, 270]:
            x = dbu * p1.x
            y = dbu * (p1.y + p2.y) / 2
        return x, y

    def get_name(c, x, y, dbu):
        s = c.begin_shapes_rec(cell.layout().layer(layer[0], layer[1]))
        while not (s.at_end()):
            if s.shape().is_text():
                label_x = s.shape().text.x * dbu
                label_y = s.shape().text.y * dbu
                if label_x == x and label_y == y:
                    return s.shape().text.string
            s.next()

    ports = []
    s = cell.begin_shapes_rec(cell.layout().layer(layer[0], layer[1]))
    while not (s.at_end()):
        if s.shape().is_path():
            width = s.shape().path_dwidth
            direction = get_direction(s.shape().path)
            # initialize Z center with none. Z center is identified in component init
            center = list(get_center(s.shape().path, cell.layout().dbu)) + [None]
            name = get_name(cell, center[0], center[1], cell.layout().dbu)
            ports.append(
                port(
                    name=name,
                    center=center,
                    width=width,
                    direction=direction,
                )
            )
        s.next()
    return ports



"""
gds_fdtd simulation toolbox.

Simulation processing module.
@author: Mustafa Hammood, 2025
"""

import numpy as np
from .core import structure, region, port, component
from .lyprocessor import (
    load_structure,
    load_region,
    load_ports,
    load_structure_from_bounds,
    dilate,
    dilate_1d,
)


def get_material(device: dict):
    """
    TODO: find a better way to handle this
    maybe use opticalmaterialspy as a universal base?
    """
    material = {'tidy3d': None, 'lum': None}
    if "tidy3d_db" in device["material"]:
        from .t3d_tools import load_material
        material['tidy3d'] = load_material(device["material"]["tidy3d_db"])

    if "lum_db" in device["material"]:
        # load material from lumerical material database, format: material model name
        if "model" in device["material"]["lum_db"]:
            mat_lum = device["material"]["lum_db"]["model"]
        material['lum'] = mat_lum

    return material


def load_component_from_tech(cell, tech, z_span=4, z_center=None):
    # load the structures in the device
    device_wg = []
    for idx, d in enumerate(tech["device"]):
        device_wg.append(
            load_structure(
                cell,
                name=f"dev_{idx}",
                layer=d["layer"],
                z_base=d["z_base"],
                z_span=d["z_span"],
                material=get_material(d),
            )
        )
    # Removing empty lists due to no structures existing in an input layer
    device_wg = [dev for dev in device_wg if dev]

    # get z_center based on structures center (minimize symmetry failures)
    if not z_center:
        z_center = np.average([d[0].z_base + d[0].z_span / 2 for d in device_wg])

    # load all the ports in the device and (optional) initialize each to have a center
    ports = load_ports(cell, layer=tech["pinrec"][0]["layer"])
    # load the device simulation region
    bounds = load_region(
        cell, layer=tech["devrec"][0]["layer"], z_center=z_center, z_span=z_span
    )

    # make the superstrate and substrate based on device bounds
    # this information isn't typically captured in a 2D layer stack
    device_super = load_structure_from_bounds(
        bounds,
        name="Superstrate",
        z_base=tech["superstrate"][0]["z_base"],
        z_span=tech["superstrate"][0]["z_span"],
        material=get_material(tech["superstrate"][0]),
    )
    device_sub = load_structure_from_bounds(
        bounds,
        name="Subtrate",
        z_base=tech["substrate"][0]["z_base"],
        z_span=tech["substrate"][0]["z_span"],
        material=get_material(tech["substrate"][0]),
    )

    # create the device by loading the structures
    return component(
        name=cell.name,
        structures=[device_sub, device_super] + device_wg,
        ports=ports,
        bounds=bounds,
    )


def from_gdsfactory(c: 'gf.Component', tech: dict, z_span: float = 4.) -> 'component':
    """Convert gdsfactory Component to a component.

    Args:
        c (gf.Component): gdsfactory component.
        tech (dict): dictionary technology stack (can be parsed from yaml) 
        z_span (float, optional): z bounds of the device (can be used to override simulation settings..). Defaults to 4..

    Returns:
        component: parsed gdsfactory component.
    """
    try:
        import gdsfactory as gf
    except ImportError:
        raise ImportError("gdsfactory is not installed. Please install it using 'pip install .[gdsfactory]'")

    device_wg = []
    ports = []

    # for each layer in the device
    for idx, layer in enumerate(c.get_polygons()):
        l = c.extract(layers={layer})       

        for i, s in enumerate(l.get_polygons()):
            name = f"poly_{idx}_{i}"
            device_wg.append(
                structure(
                    name=name,
                    polygon=l.get_polygons()[1],
                    z_base=tech["device"][idx]["z_base"],
                    z_span=tech["device"][idx]["z_span"],
                    material=get_material(tech["device"][idx]),
                    sidewall_angle=tech["device"][idx]["sidewall_angle"],
                )
            )

        # get device ports
        for p in c.ports:
            if p.layer == layer:
                z_pos = (
                    tech["device"][idx]["z_base"] + tech["device"][idx]["z_span"] / 2
                )
                ports.append(
                    port(
                        name=c.name,
                        center=list(p.center) + [z_pos],
                        width=p.width,
                        direction=p.orientation,
                    )
                )

    # get z_center based on structures center (minimize symmetry failures)
    z_center = np.average([d.z_base + d.z_span / 2 for d in device_wg])

    # expand bbox region to account for evanescent field
    def min_dim(square):
        x_dim = abs(square[0][0] - square[1][0])
        y_dim = abs(square[0][1] - square[1][1])
        if x_dim < y_dim:
            return "x"
        elif x_dim > y_dim:
            return "y"
        else:
            return "xy"

    # expand the bbox region by 1.3 um (on each side) on the smallest dimension
    bbox_poly = [[c.bbox().p1.x, c.bbox().p1.y], [c.bbox().p2.x, c.bbox().p2.y]]
    bbox = dilate_1d(bbox_poly, extension=0, dim=min_dim(bbox_poly))
    bbox_dilated = dilate(bbox, extension=1.9)
    bounds = region(vertices=bbox_dilated, z_center=z_center, z_span=z_span)

    # make the superstrate and substrate based on device bounds
    # this information isn't typically captured in a 2D layer stack
    device_super = load_structure_from_bounds(
        bounds,
        name="Superstrate",
        z_base=tech["superstrate"][0]["z_base"],
        z_span=tech["superstrate"][0]["z_span"],
        material=get_material(tech["superstrate"][0]),
    )
    device_sub = load_structure_from_bounds(
        bounds,
        name="Subtrate",
        z_base=tech["substrate"][0]["z_base"],
        z_span=tech["substrate"][0]["z_span"],
        material=get_material(tech["substrate"][0]),
    )

    # create the device by loading the structures
    return component(
        name=c.name,
        structures=[device_sub, device_super] + [device_wg],
        ports=ports,
        bounds=bounds,
    )



"""
gds_fdtd simulation toolbox.

Lumerical tools interface module.
@author: Mustafa Hammood, 2025
"""

from gds_fdtd.core import structure, component
import logging
import lumapi
import numpy as np

m_to_um = 1e-6


def structure_to_lum_poly(
    s: structure,
    lum: lumapi.FDTD,
    alpha: float = 1.0,
    group: bool = False,
    group_name: str = "group",
):
    """import a structure objecto to a lumerical instance.

    Args:
        s (structure): structure to instantiate
        lum (lumapi.FDTD): lumerical instance
        alpha (float, optional): transperancy setting. Defaults to 1..
        group (bool, optional): flag to add the structure to a given group. Defaults to False.
        group_name (str, optional): group name, if group is True. Defaults to 'group'.
    """
    if s.z_span < 0:
        bounds = (s.z_base + s.z_span, s.z_base)
    else:
        bounds = (s.z_base, s.z_base + s.z_span)

    poly = lum.addpoly(
        vertices=m_to_um * np.array(s.polygon),
        x=0,
        y=0,
        z_min=m_to_um * bounds[0],
        z_max=m_to_um * bounds[1],
        name=s.name,
        material=s.material["lum"] if isinstance(s.material, dict) else s.material,
        alpha=alpha,
    )

    if group:
        lum.addtogroup(group_name)
    lum.eval(f"?'Polygons {s.name} added';")

    return poly


def to_lumerical(c: component, lum: lumapi.FDTD, buffer: float = 2.0) -> list:
    """Add an input component with a given tech to a lumerical instance.

    Args:
        c (component): input component.
        lum (lumapi.FDTD): lumerical FDTD instance.
    """

    # TODO fix box tox handling here
    structures = []
    for s in c.structures:
        # if structure is a list then its a device (could have multiple polygons inside)
        if type(s) == list:
            for i in s:
                structures.append(
                    structure_to_lum_poly(s=i, lum=lum, group=True, group_name="device")
                )

        # if structure is not a list then its a region
        else:
            structures.append(structure_to_lum_poly(s=s, lum=lum, alpha=0.5))

    # extend ports beyond sim region
    for p in c.ports:
        structures.append(
            lum.addpoly(
                vertices=m_to_um * np.array(p.polygon_extension(buffer=buffer)),
                x=0,
                y=0,
                z_min=m_to_um * (p.center[2] - p.height / 2),
                z_max=m_to_um * (p.center[2] + p.height / 2),
                name=p.name,
                material=(
                    p.material["lum"] if isinstance(p.material, dict) else p.material
                ),
            )
        )

        lum.addtogroup("ports")
        lum.eval(f"?'port {p.name} added';")

    return structures


def make_sim_lum(
    c: component,
    lum: lumapi.FDTD,
    wavl_min: float = 1.45,
    wavl_max: float = 1.65,
    wavl_pts: int = 101,
    width_ports: float = 3.0,
    depth_ports: float = 2.0,
    symmetry: tuple[int, int, int] = (0, 0, 0),
    pol: str = "TE",
    num_modes: int = 1,
    boundary: str = "pml",
    mesh: int = 2,
    run_time_factor: float = 50,
    z_span: float | None = None,
    field_monitor_axis: str | None = None,
    visualize: bool = True,
    export_plot_file: str = 'sparam.png', 
    gpu: bool = False,
) -> dict[str, list[float]]:

    # send component to lumerical instance
    to_lumerical(c=c, lum=lum)

    if z_span == None:
        sim_size = 1e-6 * np.array([c.bounds.x_span, c.bounds.y_span, c.bounds.z_span])
    else:
        sim_size = 1e-6 * np.array([c.bounds.x_span, c.bounds.y_span, z_span])

    run_time = run_time_factor * max(sim_size) / 3e8  # 85/fwidth  # sim. time in secs

    lum.addfdtd(
        x=c.bounds.x_center * 1e-6,
        y=c.bounds.y_center * 1e-6,
        z=c.bounds.z_center * 1e-6,
        x_span=sim_size[0],
        y_span=sim_size[1],
        z_span=sim_size[2],
        simulation_time=run_time,
        mesh_accuracy=mesh,
        x_min_bc=boundary,
        y_min_bc=boundary,
        z_min_bc=boundary,
        x_max_bc=boundary,
        y_max_bc=boundary,
        z_max_bc=boundary,
    )

    if gpu:
        lum.setnamed("FDTD", "express mode", True)  # for GPU acceleration
    else:
        lum.setnamed("FDTD", "express mode", False)

    lum.setglobalsource("wavelength start", wavl_min*1e-6)
    lum.setglobalsource("wavelength stop", wavl_max*1e-6)
    lum.setglobalmonitor("frequency points", wavl_pts)

    if pol.lower() == "te":
        polarization = "fundamental TE mode"
    elif pol.lower() == "tm":
        polarization = "fundamental TM mode"
    else:
        raise ValueError(f"Polarization {pol} not supported")

    buffer = 1e-6
    for p in c.ports:
        port = lum.addport()
        lum.set("name", p.name)

        if p.direction in [90, 270]:
            if p.direction == 90:
                lum.set("y", p.center[1]*1e-6 + buffer)
                lum.set("direction", "Backward")
            else:
                lum.set("y", p.center[1]*1e-6 - buffer)
                lum.set("direction", "Forward")

            lum.set("z", p.center[2]*1e-6)
            lum.set("x", p.center[0]*1e-6)
            lum.set("injection axis", "y-axis")
            lum.set("x span", width_ports*1e-6)
            lum.set("z span", depth_ports*1e-6)
        elif p.direction in [180, 0]:
            if p.direction == 180:
                lum.set("x", p.center[0]*1e-6 - buffer)
                lum.set("direction", "Forward")
            else:
                lum.set("x", p.center[0]*1e-6 + buffer)
                lum.set("direction", "Backward")

            lum.set("y", p.center[1]*1e-6)
            lum.set("z", p.center[2]*1e-6)
            lum.set("injection axis", "x-axis")
            lum.set("y span", width_ports*1e-6)
            lum.set("z span", depth_ports*1e-6)
        else:
            raise ValueError(f"Port direction {p.direction} not supported")

        lum.set("mode selection", polarization)
        lum.set("number of field profile samples", num_modes)

    lum.save(f"{c.name}.fsp")
    lum.addsweep(3)
    lum.setsweep("s-parameter sweep", "name", "sparams")
    #input("Press Enter to continue...")
    if gpu:
        lum.runsweep("sparams", "GPU")
    else:
        lum.runsweep("sparams")

    lum.exportsweep("sparams", f"{c.name}.dat")
    sparams_sweep = lum.getsweepresult("sparams", "S parameters")

    sparams = {}
    sparams["wavl"] = [i[0] for i in sparams_sweep["lambda"]]

    ports = [port.name for port in c.ports]  # Extract port names (e.g., ['opt2', 'opt1'])
    for p1 in ports:
        for p2 in ports:
            key = f"S{ports.index(p1) + 1}{ports.index(p2) + 1}"  # Generates keys like "s11", "s12", "s21", "s22"
            sparams[f"s{ports.index(p1) + 1}_{ports.index(p2) + 1}"] = sparams_sweep[key]  # Maps to corresponding S-parameter value
    
    if visualize or export_plot_file:
        import matplotlib.pyplot as plt
        plt.figure(figsize=(10, 6))
        for param, values in sparams.items():
            if param == "wavl":
                continue  # Skip the wavelength key
            # Power in dB: 10 * log10(|S|^2)
            power_db = 10 * np.log10(np.abs(values) ** 2)
            
            plt.plot(1e6*np.array(sparams["wavl"]), power_db, label=f"|{param}|² (dB)")
        
        plt.ylabel("Transmission [dB]")
        plt.xlabel("Wavelength [μm]")
        plt.title("S-Parameters Power Response")
        plt.grid(True)
        plt.legend(loc="best")
        plt.tight_layout()
        if export_plot_file:
            plt.savefig(export_plot_file) # Saves the plot 
        if visualize:
            plt.show()

    return sparams


"""
gds_fdtd simulation toolbox.

Tidy3d interface module.
@author: Mustafa Hammood, 2025
"""

import os
import logging
import numpy as np
import matplotlib.pyplot as plt
import tidy3d as td
from .core import s_parameters, sparam, port

def load_material(material_db):
    # load material from tidy3d material database, format [material, model]
    if "model" in material_db:
        mat_tidy3d = td.material_library[material_db["model"][0]][material_db["model"][1]]
    # load tidy3d constant index material, format: refractive index
    elif "nk" in material_db:
        mat_tidy3d = td.Medium(permittivity=material_db["nk"] ** 2)
    return mat_tidy3d
    
def make_source(
    port: port,
    num_modes: int = 1,
    mode_index: int = 0,
    width: float = 3.0,
    depth: float = 2.0,
    freq0: float = 2e14,
    num_freqs: int = 5,
    fwidth: float = 1e13,
    buffer: float = -0.2,
):
    """Create a simulation mode source on an input port.

    Args:
        port (port object): Port to add source to.
        mode_index (int, optional): Mode index to launch. Defaults to 0.
        num_modes (int, optional): Number of modes to launch in the source. Defaults to 1.
        width (int, optional): Source width. Defaults to 3 microns.
        depth (int, optional): Source depth. Defaults to 2 microns.
        freq0 (_type_, optional): Source's centre frequency. Defaults to 2e14 hz.
        num_freqs (int, optional): Frequency evaluation mode ports. Defaults to 5.
        fwidth (_type_, optional): Source's bandwidth. Defaults to 1e13 hz.
        buffer (float, optional): Distance between edge of simulation and source. Defaults to 0.1 microns.

    Returns:
        td.ModeSource: Generated source.
    """

    if port.direction == 0:
        x_buffer = -buffer
        y_buffer = 0
        size = [0, width, depth]
    elif port.direction == 180:
        x_buffer = buffer
        y_buffer = 0
        size = [0, width, depth]
    elif port.direction == 90:
        x_buffer = 0
        y_buffer = -buffer
        size = [width, 0, depth]
    elif port.direction == 270:
        x_buffer = 0
        y_buffer = buffer
        size = [width, 0, depth]
    if port.direction in [180, 270]:
        direction = "+"
    else:
        direction = "-"
    msource = td.ModeSource(
        center=[port.x + x_buffer, port.y + y_buffer, port.z],
        size=size,
        direction=direction,
        source_time=td.GaussianPulse(freq0=freq0, fwidth=fwidth),
        mode_spec=td.ModeSpec(num_modes=num_modes),
        mode_index=mode_index,
        num_freqs=num_freqs,
        name=f"msource_{port.name}_idx{mode_index}",
    )
    return msource


def make_structures(device, buffer: float=4.):
    """Create a tidy3d structure object from a device objcet.

    Args:
        device (device object): Device to create the structure from.
        buffer (int, optional): Extension of ports beyond simulation region . Defaults to 2 microns.

    Returns:
        list: list of structures generated from the device.
    """
    # TODO find a better way to handle material..
    
    # TODO fix box tox handling here
    structures = []
    for s in device.structures:
        if type(s) == list:
            for i in s:
                if i.z_span < 0:
                    bounds = (i.z_base + i.z_span, i.z_base)
                else:
                    bounds = (i.z_base, i.z_base + i.z_span)
                structures.append(
                    td.Structure(
                        geometry=td.PolySlab(
                            vertices=i.polygon,
                            slab_bounds=bounds,
                            axis=2,
                            sidewall_angle=(90 - i.sidewall_angle) * (np.pi / 180),
                        ),
                        medium=i.material["tidy3d"] if isinstance(i.material, dict) else i.material,
                        name=i.name,
                    )
                )
        else:
            if s.z_span < 0:
                bounds = (s.z_base + s.z_span, s.z_base)
            else:
                bounds = (s.z_base, s.z_base + s.z_span)
            print(s.material["tidy3d"] if isinstance(s.material, dict) else s.material)
            structures.append(
                td.Structure(
                    geometry=td.PolySlab(
                        vertices=s.polygon,
                        slab_bounds=bounds,
                        axis=2,
                        sidewall_angle=(90 - s.sidewall_angle) * (np.pi / 180),
                    ),
                    medium=s.material["tidy3d"] if isinstance(s.material, dict) else s.material,
                    name=s.name,
                )
            )

    # extend ports beyond sim region
    for p in device.ports:
        structures.append(
            td.Structure(
                geometry=td.PolySlab(
                    vertices=p.polygon_extension(buffer=buffer),
                    slab_bounds=(
                        p.center[2] - p.height / 2,
                        p.center[2] + p.height / 2,
                    ),
                    axis=2,
                    sidewall_angle=(90 - device.structures[0].sidewall_angle)
                    * (np.pi / 180),
                ),
                medium=p.material["tidy3d"] if isinstance(p.material, dict) else p.material,
                name=f"port_{p.name}",
            )
        )
    return structures


def make_port_monitor(port, freqs=2e14, num_modes=1, buffer=-0.1, depth=2, width=3):
    """
    Create mode monitor object for a given port.

    Args:
        port (port object): Port to create the monitor for.
        freqs (float, optional): Mode monitor's central frequency. Defaults to 2e14 hz.
        num_modes (int, optional): Number of modes to be captured. Defaults to 1.
        buffer (float, optional): Distance between monitor and port location. Defaults to 0.2 microns.
        depth (int, optional): Monitor's depth. Defaults to 2 microns.
        width (int, optional): Monitors width. Defaults to 3 microns.

    Returns:
        monitor: Generated ModeMonitor object.
    """

    if port.direction == 0:
        x_buffer = -buffer
        y_buffer = 0
        size = [0, width, depth]
    elif port.direction == 180:
        x_buffer = buffer
        y_buffer = 0
        size = [0, width, depth]
    elif port.direction == 90:
        x_buffer = 0
        y_buffer = -buffer
        size = [width, 0, depth]
    elif port.direction == 270:
        x_buffer = 0
        y_buffer = buffer
        size = [width, 0, depth]
    # mode monitor
    monitor = td.ModeMonitor(
        center=[port.x + x_buffer, port.y + y_buffer, port.z],
        size=size,
        freqs=freqs,
        mode_spec=td.ModeSpec(num_modes=num_modes),
        name=port.name,
    )

    return monitor


def make_field_monitor(device, freqs=2e14, axis="z", z_center=None):
    """Make a field monitor for an input device

    Args:
        device (device object): Device to create field monitor for.
        freqs (float, optional): _description_. Defaults to 2e14.
        z_center (float, optional): Z center for field monitor. Defaults to None.
        axis (string, optional): Field monitor's axis. Valid options are 'x', 'y', 'z' Defaults to 'z'.

    Returns:
        FieldMonitor: Generated Tidy3D field monitor object
    """

    # identify a device field z_center if None
    if z_center is None:
        z_center = []
        for s in device.structures:
            if type(s) == list:  # i identify non sub/superstrate if s is a list
                s = s[0]
                z_center.append(s.z_base + s.z_span / 2)
        z_center = np.average(z_center)
    if axis == "z":
        center = [0, 0, z_center]
        size = [td.inf, td.inf, 0]
    elif axis == "y":
        center = [0, 0, z_center]
        size = [td.inf, 0, td.inf]
    elif axis == "x":
        center = [0, 0, z_center]
        size = [0, td.inf, td.inf]
    else:
        Exception("Invalid axis for field monitor. Valid selections are 'x', 'y', 'z'.")
    return td.FieldMonitor(
        center=center,
        size=size,
        freqs=freqs,
        name=f"{axis}_field",
    )

class sim_tidy3d:
    def __init__(
        self, in_port, device, wavl_min=1.45, wavl_max=1.65, wavl_pts=101, sim_jobs=None
    ):
        self.in_port = in_port
        self.device = device
        self.wavl_min = wavl_min
        self.wavl_max = wavl_max
        self.wavl_pts = wavl_pts
        self.sim_jobs = sim_jobs
        self.results = None

    def upload(self):
        from tidy3d import web

        # divide between job and sim, how to attach them?
        for sim_job in self.sim_jobs:
            sim = sim_job["sim"]
            name = sim_job["name"]
            sim_job["job"] = web.Job(simulation=sim, task_name=name)

    def execute(self):

        def get_directions(ports):
            directions = []
            for p in ports:
                if p.direction in [0, 90]:
                    directions.append("+")
                else:
                    directions.append("-")
            return tuple(directions)

        def get_source_direction(port):
            if port.direction in [0, 90]:
                return "-"
            else:
                return "+"

        def get_port_name(port):
            return [int(i) for i in port if i.isdigit()][0]

        def measure_transmission(in_port: port, in_mode_idx: int, out_mode_idx: int):
            """
            Constructs a "row" of the scattering matrix.
            """
            num_ports = np.size(self.device.ports)

            if isinstance(self.results, list):
                if len(self.results) == 1:
                    results = self.results[0]
                else:
                    # TBD: Handle the case where self.results is a list with more than one item
                    logging.warning(
                        "Multiple results handler is WIP, using first results entry"
                    )
                    results = self.results[-1]
            else:
                results = self.results

            input_amp = results[in_port.name].amps.sel(
                direction=get_source_direction(in_port),
                mode_index=in_mode_idx,
            )
            amps = np.zeros((num_ports, self.wavl_pts), dtype=complex)
            directions = get_directions(self.device.ports)
            for i, (monitor, direction) in enumerate(
                zip(results.simulation.monitors[:num_ports], directions)
            ):
                amp = results[monitor.name].amps.sel(
                    direction=direction, mode_index=out_mode_idx
                )
                amp_normalized = amp / input_amp
                amps[i] = np.squeeze(amp_normalized.values)

            return amps

        self.s_parameters = s_parameters()  # initialize empty s parameters

        self.results = []
        for sim_job in self.sim_jobs:
            if not os.path.exists(self.device.name):
                os.makedirs(self.device.name)
            self.results.append(
                sim_job["job"].run(
                    path=os.path.join(self.device.name, f"{sim_job['name']}.hdf5")
                )
            )
            for mode in range(sim_job["num_modes"]):
                amps_arms = measure_transmission(
                    in_port=sim_job["in_port"],
                    in_mode_idx=sim_job["source"].mode_index,
                    out_mode_idx=mode,
                )

                logging.info("Mode amplitudes in each port: \n")
                wavl = np.linspace(self.wavl_min, self.wavl_max, self.wavl_pts)
                for amp, monitor in zip(
                    amps_arms, self.results[-1].simulation.monitors
                ):
                    logging.info(f'\tmonitor     = "{monitor.name}"')
                    logging.info(f"\tamplitude^2 = {[abs(i)**2 for i in amp]}")
                    logging.info(
                        f"\tphase       = {[np.angle(i)**2 for i in amp]} (rad)\n"
                    )

                    self.s_parameters.add_param(
                        sparam(
                            idx_in=sim_job["in_port"].idx,
                            idx_out=get_port_name(monitor.name),
                            mode_in=sim_job["source"].mode_index,
                            mode_out=mode,
                            freq=td.C_0 / (wavl),
                            s=amp,
                        )
                    )
        if isinstance(self.results, list) and len(self.results) == 1:
            self.results = self.results[0]

    def _plot_any_axis(self, job_result, freq):
        """Try x/y/z until one works, put axis name in title."""
        for axis in ("x", "y", "z"):
            try:
                fig, ax = plt.subplots(1, 1, figsize=(16, 3))
                job_result.plot_field(f"{axis}_field", "Ey", freq=freq, ax=ax)
                ax.set_title(f"**{axis.upper()}‑field**")      # highlight chosen axis
                fig.show()
            except Exception:
                plt.close(fig)
        # nothing matched → silently skip


    def visualize_results(self):
        self.s_parameters.plot()
        freq = td.C_0 / ((self.wavl_max + self.wavl_min) / 2)

        if isinstance(self.results, list):
            for job_result in self.results:
                self._plot_any_axis(job_result, freq)
        else:
            self._plot_any_axis(self.results, freq)

def make_t3d_sim(
    device,
    wavl_min: float = 1.45,
    wavl_max: float = 1.65,
    wavl_pts: int = 101,
    width_ports: float = 3.0,
    depth_ports: float = 2.0,
    symmetry: tuple[int, int, int] = (0, 0, 0),
    num_freqs: int = 5,
    in_port: port | str | None = None,
    mode_index: list | int = 0,
    num_modes: int = 1,
    boundary: td.BoundarySpec = td.BoundarySpec.all_sides(boundary=td.PML()),
    grid_cells_per_wvl: int = 15,
    run_time_factor: float = 50,
    z_span: float | None = None,
    field_monitor_axis: str | None = None,
    visualize: bool = True,
):
    """Generate a single port excitation simulation.

    Args:
        device (device object): Device to simulate.
        wavl_min (float, optional): Start wavelength. Defaults to 1.45 microns.
        wavl_max (float, optional): End wavelength. Defaults to 1.65 microns.
        wavl_pts (int, optional): Number of wavelength evaluation pts. Defaults to 101.
        width_ports (int, optional): Width of source and monitors. Defaults to 3 microns.
        depth_ports (int, optional): Depth of source and monitors. Defaults to 2 microns.
        symmetry (tuple, optional): Enforcing symmetry along axes. Defaults to (0, 0, 0).
        num_freqs (int, optional): Number of source's frequency mode evaluation pts. Defaults to 5 microns.
        in_port (port object, optional): Input port. Defaults to None.
        mode_index(list, optional): Mode index to inject in source. Defaults to [0].
        num_modes(int, optional): Number of source's and monitors modes. Defaults to 1.
        boundary (td.BonudarySpec object, optional): Configure boundary conditions. Defaults to td.BoundarySpec.all_sides(boundary=td.PML()).
        grid_cells_per_wvl (int, optional): Mesh settings, grid cells per wavelength. Defaults to 15.
        run_time_factor (int, optional): Runtime multiplier factor. Set larger if runtime is insufficient. Defaults to 50.
        z_span (float, optional): Simulation's depth. Defaults to None.
        field_monitor_axis (str, optional): Flag to create a field monitor. Options are 'x', 'y', 'z', or none. Defaults to None.
        visualize (bool, optional): Simulation visualization flag. Defaults to True.

    Returns:
        simulation: Generated simulation instance.
    """

    # if no input port defined, use first as default
    if in_port is None:
        in_port = [device.ports[0]]
    if not isinstance(in_port, list):
        in_port = [in_port]
    if in_port == "all":
        in_port = device.ports[0]

    if not isinstance(mode_index, list):
        mode_index = [mode_index]

    lda0 = (wavl_max + wavl_min) / 2
    freq0 = td.C_0 / lda0
    freqs = td.C_0 / np.linspace(wavl_min, wavl_max, wavl_pts)
    fwidth = 0.5 * (np.max(freqs) - np.min(freqs))

    # define structures from device
    structures = make_structures(device)

    # define monitors
    monitors = []
    for p in device.ports:
        monitors.append(
            make_port_monitor(
                p,
                freqs=freqs,
                depth=depth_ports,
                width=width_ports,
                num_modes=num_modes,
            )
        )

    # make field monitor
    # TODO: handle mode index cases in making field monitor
    if field_monitor_axis is not None:
        monitors.append(
            make_field_monitor(device, freqs=freqs, axis=field_monitor_axis)
        )
    # simulation domain size (in microns)
    if z_span == None:
        sim_size = [device.bounds.x_span, device.bounds.y_span, device.bounds.z_span]
    else:
        sim_size = [device.bounds.x_span, device.bounds.y_span, z_span]
    run_time = (
        run_time_factor * max(sim_size) / td.C_0
    )  # 85/fwidth  # sim. time in secs

    """
    define sim jobs: create source on a given port, for each mode index
    i.e., TE and TM mode indices on 3 input ports would result in 6 sources (simulation jobs)
    some thoughts:
    when jobs are defined and sent to api, we lose the link between the simulation object and tidy3d simulation
    we maintain some data by turning the job into a dictionary object
    TODO: in the future, sim_job should be an object, and simulation should be a partial object of sim_job
    """
    sim_jobs = []
    for m in mode_index:
        for p in in_port:
            source = make_source(
                port=p,
                depth=depth_ports,
                width=width_ports,
                freq0=freq0,
                num_freqs=num_freqs,
                fwidth=fwidth,
                num_modes=num_modes,
                mode_index=m,
            )
            sim = {}
            sim["name"] = f"{device.name}_{p.name}_idx{m}"
            sim["source"] = source
            sim["in_port"] = p
            sim["num_modes"] = num_modes
            sim["sim"] = td.Simulation(
                size=sim_size,
                grid_spec=td.GridSpec.auto(
                    min_steps_per_wvl=grid_cells_per_wvl, wavelength=lda0
                ),
                structures=structures,
                sources=[source],
                monitors=monitors,
                run_time=run_time,
                boundary_spec=boundary,
                center=(
                    device.bounds.x_center,
                    device.bounds.y_center,
                    device.bounds.z_center,
                ),
                symmetry=symmetry,
            )
            sim_jobs.append(sim)

    # initialize the simulation
    simulation = sim_tidy3d(
        in_port=in_port,
        wavl_max=wavl_max,
        wavl_min=wavl_min,
        wavl_pts=wavl_pts,
        device=device,
        sim_jobs=sim_jobs,
    )

    if visualize:
        for sim_job in simulation.sim_jobs:
            sim = sim_job["sim"]
            for m in sim.monitors:
                m.help()

        source.source_time.plot(np.linspace(0, run_time, 1001))
        plt.show()

        # visualize geometry
        for sim_job in simulation.sim_jobs:
            sim = sim_job["sim"]
            gridsize = (3, 2)
            fig = plt.figure(figsize=(12, 8))
            ax1 = plt.subplot2grid(gridsize, (0, 0), colspan=2, rowspan=2)
            ax2 = plt.subplot2grid(gridsize, (2, 0))
            ax3 = plt.subplot2grid(gridsize, (2, 1))

            sim.plot(z=device.bounds.z_center, ax=ax1)
            sim.plot(x=0.0, ax=ax2)
            sim.plot(x=device.bounds.x_center, ax=ax3)
            ax1.set_title(sim_job["name"])
            plt.show()
    return simulation


"""
gds_fdtd simulation toolbox.

Core objects module.
@author: Mustafa Hammood, 2025
"""

import pya
import logging
import numpy as np
import matplotlib.pyplot as plt
from shapely.geometry import Point
from shapely.geometry.polygon import Polygon
import yaml

c0_um = 299792458000000.0  # speed of light in um/s

def is_point_inside_polygon(
    point: list[float, float], polygon_points: list[list[float, float]]
) -> bool:
    """Test if a point inside a polygon using Shapely.

    Args:
        point (list): Point for test [x, y]
        polygon_points (list): List of points defining a polygon [[x1, y1], [x2,y2], ..]

    Returns:
        bool: Test result.
    """

    # Create a Shapely Point object for the given coordinate
    point = Point(point)

    # Create a Shapely Polygon object from the list of polygon points
    polygon = Polygon(polygon_points)

    # Check if the point is inside the polygon
    return point.within(polygon) or polygon.touches(point)


class layout:
    def __init__(self, name: str, ly: pya.Layout, cell: pya.Cell):
        self.name = name
        self.ly = ly
        self.cell = cell

    @property
    def dbu(self) -> float:
        return self.ly.dbu


def calculate_polygon_extension(
    center: list[float, float], width: float, direction: float, buffer: float = 4.0
) -> list[list[float, float]]:
    """
    Calculate the polygon extension for a port.

    Args:
        center (list[float, float]): Center of the port [x, y]. Convention is in microns.
        width (float): Width of the port. Convention is in microns.
        direction (float): Direction of the port in degrees. Convention is in degrees.
        buffer (float): Buffer distance from the port. Convention is in microns.

    Returns:
        list[list[float, float]]: Polygon extension
    """
    if direction == 0:
        return [
            [center[0], center[1] + width / 2],
            [center[0] + buffer, center[1] + width / 2],
            [center[0] + buffer, center[1] - width / 2],
            [center[0], center[1] - width / 2],
        ]
    elif direction == 180:
        return [
            [center[0], center[1] + width / 2],
            [center[0] - buffer, center[1] + width / 2],
            [center[0] - buffer, center[1] - width / 2],
            [center[0], center[1] - width / 2],
        ]
    elif direction == 90:
        return [
            [center[0] - width / 2, center[1]],
            [center[0] - width / 2, center[1] + buffer],
            [center[0] + width / 2, center[1] + buffer],
            [center[0] + width / 2, center[1]],
        ]
    elif direction == 270:
        return [
            [center[0] - width / 2, center[1]],
            [center[0] - width / 2, center[1] - buffer],
            [center[0] + width / 2, center[1] - buffer],
            [center[0] + width / 2, center[1]],
        ]


class port:
    """
    Represents an optical port object in a component.

    A port defines a connection point with properties like position, width, and direction.

    Attributes:
        name (str): Name of the port, typically containing a numeric identifier.
        center (list[float, float, float]): 3D coordinates of the port center [x, y, z]. Convention is in microns.
        width (float): Width of the port. Convention is in microns.
        direction (float): Direction of the port. Convention is in degrees. Directions supported are 0, 90, 180, 270.
        height (float): Height of the port, assigned during component initialization. Convention is in microns.
        material (str): Material of the port, assigned during component initialization.
    """

    def __init__(
        self,
        name: str,
        center: list[float, float, float],
        width: float,
        direction: float,
    ):
        """
        Initialize a port object.

        Args:
            name (str): Name of the port, typically containing a numeric identifier.
            center (list[float, float, float]): 3D coordinates of the port center [x, y, z].
            width (float): Width of the port in microns.
            direction (float): Direction of the port in degrees.
        """
        self.name = name
        self.center = center
        self.width = width
        self.direction = direction
        # initialize height as none
        # will be assigned upon component __init__
        # TODO: feels like a better way to do this..
        self.height = None
        self.material = None

        if self.direction not in [0, 90, 180, 270]:
            raise ValueError(
                f"Invalid direction: {self.direction}. Supported directions are 0, 90, 180, 270."
            )

    @property
    def x(self) -> float:
        """
        Get the x-coordinate of the port center.

        Returns:
            float: x-coordinate in microns.
        """
        return self.center[0]

    @property
    def y(self) -> float:
        """
        Get the y-coordinate of the port center.

        Returns:
            float: y-coordinate in microns.
        """
        return self.center[1]

    @property
    def z(self) -> float:
        """
        Get the z-coordinate of the port center.

        Returns:
            float: z-coordinate in microns.
        """
        return self.center[2]

    @property
    def idx(self) -> int:
        """
        Extract the index of the port from its name.

        The index is extracted by taking the digits in the name in reverse order.
        For example, "port42" would yield an index of 24.

        Returns:
            int: The extracted port index.
        """
        return int("".join(char for char in reversed(self.name) if char.isdigit()))

    def polygon_extension(self, buffer: float = 4.0) -> list[list[float, float]]:
        """
        Calculate the polygon extension for this port.

        This creates a rectangular polygon extending from the port in the direction
        specified by the port's direction attribute.

        Args:
            buffer (float, optional): Buffer distance from the port in microns. Defaults to 4.0.

        Returns:
            list[list[float, float]]: Polygon extension as a list of [x,y] coordinates.
        """
        return calculate_polygon_extension(
            self.center, self.width, self.direction, buffer
        )


class structure:
    """
    Represents a physical structure in the component with geometric and material properties.

    This class defines a 3D structure with a 2D polygon base extruded vertically,
    including material properties and sidewall angle for fabrication realism.
    """

    def __init__(
        self,
        name: str,
        polygon: list[list[float, float]],
        z_base: float,
        z_span: float,
        material: str,
        sidewall_angle: float = 90.0,
    ):
        """
        Initialize a structure with geometric and material properties.

        Args:
            name (str): Unique identifier for the structure.
            polygon (list[list[float, float]]): 2D polygon defining the structure's horizontal cross-section,
                                               formatted as [[x1,y1], [x2,y2], ...].
            z_base (float): Base z-coordinate in microns where the structure begins.
            z_span (float): Vertical height/thickness of the structure in microns.
            material (str): Material identifier for the structure.
            sidewall_angle (float, optional): Angle of the sidewalls in degrees, where 90.0 means vertical walls.
                                             Defaults to 90.0.
        """
        self.name = name
        self.polygon = polygon  # polygon should be in the form of list of list of 2 pts, i.e. [[0,0],[0,1],[1,1]]
        self.z_base = z_base
        self.z_span = z_span
        self.material = material
        self.sidewall_angle = sidewall_angle


class region:
    """
    Represents a 3D region defined by a 2D polygon and vertical extent.
    
    This class defines a region with vertices in the x-y plane and a vertical
    extent defined by z_center and z_span.
    """
    
    def __init__(self, vertices: list[list[float]], z_center: float, z_span: float):
        """
        Initialize a region with vertices and vertical dimensions.
        
        Args:
            vertices (list[list[float]]): List of [x,y] coordinates defining the region's polygon.
            z_center (float): Center z-coordinate of the region in microns.
            z_span (float): Vertical extent/thickness of the region in microns.
        """
        self.vertices = vertices
        self.z_center = z_center
        self.z_span = z_span

    @property
    def x(self) -> list[float]:
        """
        Get all x-coordinates of the vertices.
        
        Returns:
            list[float]: List of x-coordinates.
        """
        return [i[0] for i in self.vertices]

    @property
    def y(self) -> list[float]:
        """
        Get all y-coordinates of the vertices.
        
        Returns:
            list[float]: List of y-coordinates.
        """
        return [i[1] for i in self.vertices]

    @property
    def x_span(self) -> float:
        """
        Calculate the span (width) of the region in the x-direction.
        
        Returns:
            float: Width of the region in microns.
        """
        return abs(min(self.x) - max(self.x))

    @property
    def y_span(self) -> float:
        """
        Calculate the span (height) of the region in the y-direction.
        
        Returns:
            float: Height of the region in microns.
        """
        return abs(min(self.y) - max(self.y))

    @property
    def x_center(self) -> float:
        """
        Calculate the center x-coordinate of the region.
        
        Returns:
            float: Center x-coordinate in microns.
        """
        return (min(self.x) + max(self.x)) / 2

    @property
    def y_center(self) -> float:
        """
        Calculate the center y-coordinate of the region.
        
        Returns:
            float: Center y-coordinate in microns.
        """
        return (min(self.y) + max(self.y)) / 2

    @property
    def x_min(self) -> float:
        """
        Get the minimum x-coordinate of the region.
        
        Returns:
            float: Minimum x-coordinate in microns.
        """
        return min(self.x)

    @property
    def x_max(self) -> float:
        """
        Get the maximum x-coordinate of the region.
        
        Returns:
            float: Maximum x-coordinate in microns.
        """
        return max(self.x)

    @property
    def y_min(self) -> float:
        """
        Get the minimum y-coordinate of the region.
        
        Returns:
            float: Minimum y-coordinate in microns.
        """
        return min(self.y)

    @property
    def y_max(self) -> float:
        """
        Get the maximum y-coordinate of the region.
        
        Returns:
            float: Maximum y-coordinate in microns.
        """
        return max(self.y)



def initialize_ports_z(ports: list["port"], structures: list["structure"]) -> None:
    """
    Initialize the z-coordinate, height, and material of ports based on their position within structures.
    
    This function checks if each port is located within any structure and sets the port's z-coordinate,
    height, and material accordingly.
    
    Args:
        ports (list["port"]): List of port objects to initialize.
        structures (list["structure"]): List of structure objects to check against.
        
    Returns:
        None
    """
    # iterate through each port
    for p in ports:
        # check if port location is within any structure
        for s in structures:
            # TODO: hack: if s is a list then it's not a box/clad region, find a better way to identify this..
            if type(s) == list:
                for poly in s:
                    if is_point_inside_polygon(p.center[:2], poly.polygon):
                        p.center[2] = s[0].z_base + s[0].z_span / 2
                        p.height = s[0].z_span
                        p.material = s[0].material
        if p.height == None:
            logging.warning(f"Cannot find height for port {p.name}")
    return
class component:
    """
    A component consisting of structures, ports, and boundaries.
    
    This class represents a complete photonic component that can be simulated
    or exported to GDS format.
    """
    
    def __init__(self, name: str, structures: list[structure], ports: list[port], bounds: list[region]):
        """
        Initialize a photonic component.
        
        Args:
            name: The name of the component.
            structures: List of structures (geometries) in the component.
            ports: List of ports for input/output connections.
            bounds: Boundaries of the component.
        """
        self.name = name
        self.structures = structures
        self.ports = ports
        self.bounds = bounds
        initialize_ports_z(ports = self.ports, structures = self.structures)  # initialize ports z center and z span

    def export_gds(self, export_dir: str = None, dbu: float = 0.001, layer: list = [1, 0]) -> None:
        """
        Export the component to a GDS file.
        
        Args:
            export_dir: Directory to export the GDS file to. Defaults to current working directory.
            dbu: Database unit in microns. Defaults to 0.001 (1 nm).
            layer: GDS layer specification as [layer_number, datatype]. Defaults to [1, 0].
        """
        import os
        import klayout.db as pya

        layout = pya.Layout()
        layout.dbu = dbu  # Set the database unit to 0.001 um
        top_cell = layout.create_cell(self.name)
        layer_info = pya.LayerInfo(layer[0], layer[1])
        layer = layout.layer(layer_info)

        for polygon in [s[0].polygon for s in self.structures if isinstance(s, list)]:
            pya_polygon = pya.Polygon(
                [
                    pya.Point(int(point[0] / layout.dbu), int(point[1] / layout.dbu))
                    for point in polygon
                ]
            )
            top_cell.shapes(layer).insert(pya_polygon)

        if export_dir is None:
            export_dir = os.getcwd()
        layout.write(os.path.join(export_dir, f"{self.name}.gds"))
        return 


class s_parameters:
    def __init__(self, entries=None):
        if entries is None:
            self._entries = []
        else:
            self._entries = entries
        return

    @property
    def S(self):
        return dict(zip([i.label for i in self._entries], self._entries))

    def add_param(self, sparam):
        self._entries.append(sparam)

    def entries_in_mode(self, mode_in=0, mode_out=0):
        entries = []
        for s in self._entries:
            if s.mode_in == mode_in and s.mode_out == mode_out:
                entries.append(s)
        return entries

    def entries_in_ports(self, input_entries=None, idx_in=0, idx_out=0):
        entries = []
        if input_entries == None:
            input_entries = self._entries

        for s in input_entries:
            if s.idx_in == idx_in and s.idx_out == idx_out:
                entries.append(s)
        return entries

    def plot(self):
        fig, ax = plt.subplots(1, 1)
        ax.set_xlabel("Wavelength [microns]")
        ax.set_ylabel("Transmission [dB]")
        for i in self._entries:
            logging.info("Mode amplitudes in each port: \n")
            mag = [10 * np.log10(abs(i) ** 2) for i in i.s]
            phase = [np.angle(i) ** 2 for i in i.s]
            ax.plot(1e6 * c0_um / i.freq, mag, label=i.label)
        ax.legend()
        fig.show()
        return fig, ax

class sparam:
    def __init__(self, idx_in, idx_out, mode_in, mode_out, freq, s):
        self.idx_in = idx_in
        self.idx_out = idx_out
        self.mode_in = mode_in
        self.mode_out = mode_out
        self.freq = freq
        self.s = s

    @property
    def label(self):
        return f"S{self.idx_out}{self.idx_in}_idx{self.mode_out}{self.mode_in}"

    def plot(self):
        fig, ax = plt.subplots(1, 1)
        ax.plot((1e-6 * c0_um) / np.array(self.freq), 10 * np.log10(self.s**2))
        ax.set_xlabel("Wavelength [um]")
        ax.set_ylabel("Transmission [dB]")
        ax.set_title("Frequency vs S")
        fig.show()
        return fig, ax

def parse_yaml_tech(file_path):
    with open(file_path, "r") as file:
        data = yaml.safe_load(file)

    technology = data.get("technology", {})
    parsed_data = {
        "name": technology.get("name", "Unknown"),
        "substrate": [],
        "superstrate": [],
        "pinrec": [],
        "devrec": [],
        "device": [],
    }

    # Parsing substrate layer
    substrate = technology.get("substrate", {})
    parsed_data["substrate"].append(
        {
            "z_base": substrate.get("z_base"),
            "z_span": substrate.get("z_span"),
            "material": substrate.get("material"),
        }
    )

    # Parsing superstrate layer
    superstrate = technology.get("superstrate", {})
    parsed_data["superstrate"].append(
        {
            "z_base": superstrate.get("z_base"),
            "z_span": superstrate.get("z_span"),
            "material": superstrate.get("material"),
        }
    )

    # Parsing pinrec layers
    parsed_data["pinrec"] = [
        {"layer": list(pinrec.get("layer"))}  # Convert to list
        for pinrec in technology.get("pinrec", [])
    ]

    # Parsing devrec layers
    parsed_data["devrec"] = [
        {"layer": list(devrec.get("layer"))}  # Convert to list
        for devrec in technology.get("devrec", [])
    ]

    # Parsing device layers
    parsed_data["device"] = [
        {
            "layer": list(device.get("layer")),  # Convert to list
            "z_base": device.get("z_base"),
            "z_span": device.get("z_span"),
            "material": device.get("material"),
            "sidewall_angle": device.get("sidewall_angle"),
        }
        for device in technology.get("device", [])
    ]

    return parsed_data
